import functools

import sentence_mixing.logic.analyze
import sentence_mixing.logic.analyze_step_3 as step_3
import sentence_mixing.logic.global_audio_data as global_audio_data
import sentence_mixing.logic.parameters as params
from sentence_mixing.model.scorable import Scorable


class Choice(Scorable):
    """
    This class represents a node in the decision tree.

    Also contains a step 3 score, rating the association comparatively to all parent associations.

    Class attributes:
    parent - father node in the tree
    association - last association decided by parent
    _audio_score - audio score of association (step 3 scoring)
    _previous_score - previous score of association (step 3 scoring)
    children - children nodes in the tree
    """

    def __init__(
        self,
        parent,
        current_association,
        nodes_left,
        audio_spectral_score,
        audio_amplitude_score,
        previous_score,
    ):
        self.parent = parent
        self.nodes_left = nodes_left
        self.children = []

        self.association = current_association

        self._audio_spectral_score = audio_spectral_score
        self._audio_amplitude_score = audio_amplitude_score
        self._previous_score = previous_score

    def get_self_and_previous_choices(self):
        """This is a generator used to retrieve the current Choice and all its parents"""

        choice = self
        while choice is not None:
            yield choice
            choice = choice.parent

    @functools.lru_cache(maxsize=None)
    def _create_children(self):
        """
        Creates choice children
        This function is decomposed in two parts:
        -skip checking
        -normal workflow

        Skip checking:
            First, it tries to spot if an homophonous word sequence matching target can be found
            from the current audio phonem.

            If a skip occurs, the function creates SkippedChoice typed children instead of regular
            Choice typed children.

            Example:
                -TargetSentence: "Les <BLANK> puissances <BLANK> de"
                -In the audio subtitles, you can find "puissance deux"
                -self.assocation = ['p', 'p']

                The function, by checking the audio words followed by audio phonem 'p', will spot
                a two consecutives homophones sequence, common in target and audio:
                "puissances de" for target and "puissance deux" for audio.

                These two sequences are considered as homophones because they contains strictly
                the same phonems. Please note that <BLANK> token is skipped while looking for
                word sequence (check the functionlogic.utils.get_sequence() for more details).

                The function will then force the choice all the phonems of "puissance de" by
                SkippedChoice children.


            Then, it applies the same logic for identical phonem sequences.
            A common sequence ends at the end of the current word.

            Example:
                -TargetSentence: "Ma <BLANK> trique <BLANK> sauce"
                -In audio subtitles, you can find the word "géométrique saucisse"
                -self.association = ['t', 't']

                The function will check if a common phonem sequence can be found between audio and
                target phonem.
                It will find a common sequence, starting from current audio and target phonem 't':
                phonems 't', 'r', 'i', 'k' ("trique").

                Warning: despite "sauce" and "saucisse" contains common phonems at the begginging
                        ('s', 'o', 's'), it will not be included in previous sequence ("trique"),
                        because it is not part of the same word !

                        Of course, when self.association will contains phonems ['s', 's'], this
                        function will detect the common phonem sequence 's', 'o', 's'.


        Normal workflow:

            If no skip opportunity was found, the function will simply try to get the best
            association for its children.
            Normal workflow is handled by logic.analyze() function. Check it for more details.
        """

        # Leaf child
        if self.association.target_phonem.next_in_seq() is None:
            return []

        # Check for word skipping
        if self.association.target_phonem.get_index_in_word() == 0:
            # Checks homophones basing on the dictionary
            dictionary_homophones_phonems = list(
                self.association.sequence_dictionary_homophones_phonems()
            )
            # Checks homophones basing on audio word's list of phonems
            aligner_homophones_phonems = list(
                self.association.sequence_aligner_homophones_phonems()
            )

            # If at least a one-word-length sequence has be found
            if len(dictionary_homophones_phonems) > 2:
                return [SkippedChoice(self, dictionary_homophones_phonems[1:])]
            # If the word sequence contains a word of only two phonem
            elif len(dictionary_homophones_phonems) == 2:
                return [
                    Choice(
                        self,
                        dictionary_homophones_phonems[1],
                        self.nodes_left,
                        *self.compute_child_step_3_score(
                            dictionary_homophones_phonems[1]
                        ),
                    )
                ]

            # If at least a one-word-length sequence has be found
            elif len(aligner_homophones_phonems) > 2:
                return [SkippedChoice(self, aligner_homophones_phonems[1:])]
            # If the word sequence contains a word of only two phonem
            elif len(aligner_homophones_phonems) == 2:
                return [
                    Choice(
                        self,
                        aligner_homophones_phonems[1],
                        self.nodes_left,
                        *self.compute_child_step_3_score(
                            aligner_homophones_phonems[1]
                        ),
                    )
                ]

        # Check for phonem skipping
        same_phonems = (
            self.association.sequence_same_phonems_first_word_truncated()
        )
        # A common phonem sequence has been found
        if len(same_phonems) > 2:
            return [SkippedChoice(self, same_phonems[1:])]
        elif len(same_phonems) == 2:
            return [
                Choice(
                    self,
                    same_phonems[1],
                    self.nodes_left,
                    *self.compute_child_step_3_score(same_phonems[1]),
                )
            ]

        # Normal workflow
        target_phonem = self.association.target_phonem.next_in_seq()
        return sentence_mixing.logic.analyze.compute_children(
            target_phonem, self.nodes_left, self
        )

    def get_combos(self):
        """Recursively creates children and returns a list of combos created from leaf nodes"""

        self.children = self._create_children()

        if len(self.children) == 0:
            return [Combo(self)]

        return sum((child.get_combos() for child in self.children), [])

    def _get_split_score(self):
        step_3_scores = {
            "step_3_audio_spectral": self._audio_spectral_score,
            "step_3_audio_amplitude": self._audio_amplitude_score,
            "step_3_same_word_previous_phonems": self._previous_score,
        }
        return {**self.association.get_split_score(), **step_3_scores}

    def compute_child_step_3_score(self, association):
        """
        Computes step 3 score for a given association.
        This score depends on the previous chosen associations.

        Assume that self has a parent.
        """

        previous_associations = []
        previous_associations.extend(
            c.association for c in self.get_self_and_previous_choices()
        )

        all_associations = [association] + previous_associations

        rate = []

        if association.target_phonem.get_type() == "VOWEL":
            rate.append(
                step_3.step_3_audio_spectral_rating(
                    step_3.get_last_vowel(previous_associations),
                    association.audio_phonem,
                )
                * params.RATING_SPECTRAL_SIMILARITY
            )
        else:
            rate.append(0)

        if association.target_phonem.transcription != "sp":
            rate.append(
                step_3.step_3_audio_amplitude_rating(
                    (previous_associations), association,
                )
                * -params.RATING_AMPLITUDE_DIFFERENCE
            )
        else:
            rate.append(0)

        rate.append(
            params.RATING_LENGTH_SAME_PHONEM
            * step_3.step_3_n_following_previous_phonems(all_associations)
        )

        return rate


class SkippedChoice(Choice):
    """
    This subclass of Choice is used when a skip is performed.
    It represent a filiform portion of the tree.

    For example, if a Choice detected a the following common phonem sequence 't', 'r', 'i', 'k',
    a sequence of 4 SkippedChoice will be created.

    Class attributes:
    _associations_list - stack of all the remaining phonems of the skipped sequence.
                         With the previous example: first SkippedChoice will have association list
                         of 'r', 'i', 'k', second will have 'i', 'k'...
    """

    def __init__(self, parent, associations_list):
        Choice.__init__(
            self,
            parent,
            associations_list[0],
            parent.nodes_left,
            *parent.compute_child_step_3_score(associations_list[0]),
        )
        self._associations_list = associations_list[1:]

    def _create_children(self):
        """
        In opposite to _create_children() method of Choice, _create_children() method of
        SkippedChoice doesn't perform any analyze but only creates a children with the next
        phonem in the skipped sequence (associations_list).

        When the skip is finished (_associations_list is empty): creates a regular Choice
        """

        if len(self._associations_list) > 1:
            return [SkippedChoice(self, self._associations_list)]
        else:
            score = self.compute_child_step_3_score(self._associations_list[0])
            return [
                Choice(
                    self, self._associations_list[0], self.nodes_left, *score
                )
            ]


class Combo:
    """Represents a complete combo"""

    def __init__(self, leaf_choice):
        self.leaf_choice = leaf_choice

    def __repr__(self):
        return f"<combo: {list(self.get_audio_phonems())} >"

    def get_score(self):
        return sum(
            ch.get_total_score()
            for ch in self.leaf_choice.get_self_and_previous_choices()
        )

    def get_audio_phonems(self):
        return [ch.association.audio_phonem for ch in self.get_choices()]

    def get_choices(self):
        return list(
            reversed(list(self.leaf_choice.get_self_and_previous_choices()))
        )



import functools

import sentence_mixing.logic.text_parser as tp


class Sequencable:
    def next_in_seq(self):
        raise NotImplementedError()

    def previous_in_seq(self):
        raise NotImplementedError()


class Sentence(Sequencable):
    """
    Represents an abstract sequence of words

    Class attributes:
    original_text - string of sentence
    words - list of associated words
    index_words - dictionary associating for each word its index in words list
    """

    def __init__(self, original_text):
        self.original_text = original_text
        self.words = []
        self.index_words = {}

    def set_words(self, words):
        self.words = list(words)
        self.index_words = {p: i for i, p in enumerate(self.words)}

    def add_word(self, word):
        self.index_words[word] = len(self.words)
        self.words.append(word)

    def get_index_word(self, word):
        assert self.index_words is not None and word in self.index_words
        return self.index_words[word]

    @functools.lru_cache(maxsize=None)
    def get_phonem_size(self):
        return sum(len(ps.phonems) for ps in self.words)

    @functools.lru_cache(maxsize=None)
    def get_phonem_index(self, phonem):
        i = phonem.get_index_in_word()
        return i + sum(
            len(ps.phonems)
            for ps in self.words[: self.get_index_word(phonem.word)]
        )


class Word(Sequencable):
    """
    Represents an abstract word

    Class attributes:
    sentence - sentence where the word comes from
    token - word as found in the dictionary
    original_word - word as found in the original sentence
    phonems - list of associated phonems
    index_phonems - dictionary associating for each phonem its index in phonems list
    """

    def __init__(self, sentence_class, sentence, token, original_word):
        self.token = token
        self.original_word = original_word
        assert type(sentence) == sentence_class
        self.sentence = sentence
        self.phonems = []
        self.index_phonems = {}

    def set_phonems(self, phonems):
        self.phonems = list(phonems)
        self.index_phonems = {p: i for i, p in enumerate(self.phonems)}

    def add_phonem(self, phonem):
        self.index_phonems[phonem] = len(self.phonems)
        self.phonems.append(phonem)

    def get_index_phonem(self, phonem):
        assert self.index_phonems is not None and phonem in self.index_phonems
        return self.index_phonems[phonem]

    def get_index_in_sentence(self):
        return self.sentence.get_index_word(self)

    def next_in_seq(self):
        if self == self.sentence.words[-1]:
            sentence = self.sentence.next_in_seq()
            while sentence is not None and len(sentence.words) == 0:
                sentence = sentence.next_in_seq()
            if sentence is None:
                return None
            else:
                return sentence.words[0]
        return self.sentence.words[self.get_index_in_sentence() + 1]

    def previous_in_seq(self):
        if self == self.sentence.words[0]:
            sentence = self.sentence.previous_in_seq()
            while sentence is not None and len(sentence.words) == 0:
                sentence = sentence.previous_in_seq()
            if sentence is None:
                return None
            else:
                return sentence.words[-1]
        return self.sentence.words[self.get_index_in_sentence() - 1]

    def has_same_phonems_transcription(self, other):
        assert isinstance(other, Word)
        return len(self.phonems) == len(other.phonems) and all(
            map(
                lambda phs: phs[0].transcription == phs[1].transcription,
                zip(self.phonems, other.phonems),
            )
        )

    def __repr__(self):
        return f"<Word: {self.token}>"


class Phonem(Sequencable):
    """
    Represents an abstract phonem

    Class attributes:
    word - word where the phonem comes from
    transcription - phonem transcription as found  in the dictionary
    """

    def __init__(self, word_class, word, transcription):
        self.transcription = transcription
        assert type(word) == word_class
        self.word = word

    def get_index_in_word(self):
        return self.word.get_index_phonem(self)

    def next_in_seq(self):
        if self == self.word.phonems[-1]:
            word = self.word.next_in_seq()
            while word is not None and len(word.phonems) == 0:
                word = word.next_in_seq()
            if word is None:
                return None
            else:
                return word.phonems[0]
        return self.word.phonems[self.get_index_in_word() + 1]

    def previous_in_seq(self):
        if self == self.word.phonems[0]:
            word = self.word.previous_in_seq()
            while word is not None and len(word.phonems) == 0:
                word = word.previous_in_seq()
            if word is None:
                return None
            else:
                return word.phonems[-1]
        return self.word.phonems[self.get_index_in_word() - 1]

    def get_type(self):
        return tp.get_consonant_vowel_dict()[self.transcription]

    def __repr__(self):
        return f"<Phonem: {self.transcription}>"



import base64
import os
import shutil
import tempfile
from itertools import groupby
from pathlib import Path

import textgrid
import youtube_dl
from scipy.io import wavfile

import sentence_mixing.config as config
import sentence_mixing.logic.text_parser as tp
import sentence_mixing.model.audio as audio
from sentence_mixing.model.video import Video


def _create_videos(video_urls):
    """Builds basic video objects by downloading videos"""

    paths = _dl_videos(video_urls)

    # TODO: modifier le "" pour mettre l'extension de la vidéo
    return [
        Video(url, base_path, subtitles_extension)
        for url, (base_path, subtitles_extension) in zip(video_urls, paths)
    ]


class Logger(object):
    def debug(self, msg):
        pass

    def warning(self, msg):
        pass

    def error(self, msg):
        print(msg)


def _dl_videos(urls):
    """
    Downloads audio and subs for videos at the given urls

    Arguments:
    - urls: iterable of youtube urls of the wanted videos

    Returns:
    A list of tuples (path, subtitle_file_extension) matching the given urls
    """

    paths = []
    for url in urls:
        ydl_opts = {
            "writesubtitles": True,
            "writeautomaticsub": True,
            "subtitleslangs": [config.get_property("lang")],
            "outtmpl": ".downloads/%(id)s.%(title)s.%(ext)s",
            "format": "bestaudio/best",
            "logger": Logger(),
            "postprocessors": [
                {
                    "key": "FFmpegExtractAudio",
                    "preferredcodec": "wav",
                    "preferredquality": "192",
                }
            ],
        }
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            filename = ydl.prepare_filename(ydl.extract_info(url))
            ydl.download([url])
            base_path = os.path.splitext(filename)[0]
            # TODO: rendre le .fr.vtt dynamique en fonction de la langue des sous-titres
            extension = ".{}.vtt".format(config.get_property("lang"))
            paths.append((base_path, extension))
    return paths


def _create_subs(videos):
    """Adds a list of basic subtitle object for each given video by parsing subtitle file"""

    subtitle_files = [(video, video.get_subtitle_file()) for video in videos]

    for video, subtitle_file in subtitle_files:
        for caption in subtitle_file:
            video.add_subtitle(_parse_caption(video, caption))


def _parse_caption(video, caption):
    """Parses a webvtt caption to an AudioSubtitle object"""

    def _get_sec(time_str):
        h, m, s = time_str.split(":")
        return float(h) * 3600 + float(m) * 60 + float(s)

    # TODO: revoir le \n
    text = tp.transform_numbers(caption.text.split("\n")[-1])
    start = _get_sec(caption.start)
    end = _get_sec(caption.end)

    return audio.SubtitleLine(text, start, end, video)


def _split_audio_in_files(video):
    """
    Saves each SubtitleLine in a separate audio file and text file in order to be processed by
    Montreal aligner
    """

    folder_name = config.get_property("folder")
    shutil.rmtree(folder_name, True)
    Path(folder_name).mkdir(parents=True, exist_ok=True)

    video_hashed_name = video.get_hashed_basename()

    rate, data = video.get_audio_wave()

    for i, sub in enumerate(video.subtitles):
        sub_base_name = os.path.join(folder_name, video_hashed_name + str(i))

        start_frame = int(sub.start * rate)
        end_frame = int(sub.end * rate)

        sub_audio_data = data[start_frame:end_frame]
        wavfile.write(sub_base_name + ".wav", rate, sub_audio_data)

        with open(sub_base_name + ".lab", "w", encoding="utf-8") as subfile:
            subfile.write(sub.original_text)


def _align_phonems():
    """
    Launches the aligner

    Returns the temporary directory containing the analysis results
    """

    folder = config.get_property("folder")
    speakers = 1

    align_exe = config.get_property("align_exe")
    dict_path = config.get_property("dict_path")
    trained_model = config.get_property("trained_model")
    out_dir = tempfile.mkdtemp()
    temp_dir = tempfile.mkdtemp()

    command = f'{align_exe} "{folder}" "{dict_path}" "{trained_model}" "{out_dir}" -t "{temp_dir}" -s {str(speakers)}'

    ret = os.system(command)
    assert ret == 0

    return out_dir


def _parse_align_result(textgrid_path, subtitle):
    """
    Retrieves and parses the result of the aligner and enriches given subtitle objet by creating
    associated AudioWords and AudioPhonem
    """

    if not os.path.exists(textgrid_path):
        return 0
    t = textgrid.TextGrid.fromFile(textgrid_path)

    words = t[0]

    phonems = t[1]
    i_phonems = 0

    for word in words:
        original_word = word.mark
        start = word.bounds()[0] + subtitle.start
        end = word.bounds()[1] + subtitle.start
        token = tp.from_word_to_token(word.mark)

        audio_word = audio.AudioWord(
            subtitle, token, original_word, start, end
        )
        subtitle.add_word(audio_word)

        while (
            i_phonems < len(phonems)
            and phonems[i_phonems].bounds()[0] < word.bounds()[1]
        ):
            # If the phonem is included in the word
            if phonems[i_phonems].bounds()[0] >= word.bounds()[0]:
                transcription = phonems[i_phonems].mark
                start_phon = phonems[i_phonems].bounds()[0] + subtitle.start
                end_phon = phonems[i_phonems].bounds()[1] + subtitle.start

                # Only creates phonem if it is exploitable
                # Excludes phonems whose transcription like 'sil', 'spn', ''
                if transcription in tp.get_all_phonems():
                    audio_word.add_phonem(
                        audio.AudioPhonem(
                            audio_word, transcription, start_phon, end_phon
                        )
                    )
            i_phonems += 1


def preprocess_and_align(video_urls):
    """
    Build all the model objects for several video urls by downloading videos and analysing the
    videos using Montreal aligner.

    Argument:
    video_urls - list containing youtube videos url

    Returns a list containing all the video objects
    """

    # Creates basic Video objects
    videos = _create_videos(video_urls)
    # TODO maybe get all the data analyzed at once (waav resampling needed)

    # Prepares the aligner
    for video in videos:
        # Enriches Video objects with SubtitleLine objects
        _create_subs([video])

        _split_audio_in_files(video)
        # Launches the aligner
        out_dir = _align_phonems()

        # Enriches Video objects with AudioWords and AudioPhonem objects
        for i, subtitle in enumerate(video.subtitles):
            textgrid_path = os.path.join(
                out_dir, video.get_hashed_basename() + str(i) + ".TextGrid"
            )
            _parse_align_result(textgrid_path, subtitle)

    return videos



import concurrent.futures
import functools
import os
import tempfile
from pathlib import Path

import Levenshtein
import numpy as np
import pocketsphinx as ps

import sentence_mixing.config as config
import sentence_mixing.logic.audio_analysis
import sentence_mixing.logic.parameters as params
import sentence_mixing.parameter_tuning.speech_to_text_dict as stt_dict
import sentence_mixing.sentence_mixer as sm
import sentence_mixing.video_creator.audio

# minimal_phonem_range = [0, 0.03, 0.1]
# maximal_consonant_length_range = [0.10, 0.25, 0.5]
# maximal_vowel_length_range = [0.10, 0.25, 0.5]
# silence_power_range = [1, 2, 4]
# start_modif_range = [0.8, 1.0, 1.2]
# rate_poer_range = [0.9, 1.1, 1.3]


# maximal_minimal_phonem_length_malus_range = [500, 1000, 2000]
# maximal_maximal_consonant_length_malus_range = [500, 1000, 2000]
# maximal_maximal_vowel_length_malus_range = [500, 1000, 2000]
# rating_length_same_phonem_range = [80, 150, 250]
# rating_length_same_word_range = [100, 170, 300]
# score_same_transcription_range = [0, 200, 400]
# score_silence_amplitude_range = [0, 200, 400]
# score_duration_rang = [0, 200, 400]
# rating_spectral_similarity_range = [0, 50, 150]
# rating_amplitude_difference_range = [-500, -250, 0]
# score_same_audio_word_range = [80, 200, 500]
# random_span_range = [0, 50, 100]

sentences = ["l'abeille ça pique", "le confinement me prend la tête"]
video_urls = ["https://www.youtube.com/watch?v=-rTBodCtBAM"]
NB_COMBOS = 5
FACTOR = 1000


@functools.lru_cache(maxsize=1)
def get_decoder():
    """Returns speech to text decoder, created with proper path to models, weights and dict"""

    config.get_property("stt_model_path")

    stt_config = ps.Decoder.default_config()
    stt_config.set_string(
        "-hmm",
        os.path.join(config.get_property("stt_model_path"), "acoustic-model"),
    )  # set the path of the hidden Markov model (HMM) parameter files
    stt_config.set_string(
        "-lm",
        os.path.join(
            config.get_property("stt_model_path"), "language-model.lm.bin"
        ),
    )
    stt_config.set_string("-dict", config.get_property("stt_tmp_dict_path"))
    stt_config.set_string(
        "-logfn", os.devnull
    )  # disable logging (logging causes unwanted output in terminal)
    decoder = ps.Decoder(stt_config)

    return decoder


def get_parameters():
    return (
        np.array(
            [
                params.MAXIMAL_MINIMAL_PHONEM_LENGH_MALUS,
                params.MAXIMAL_MAXIMAL_CONSONANT_LENGTH_MALUS,
                params.MAXIMAL_MAXIMAL_VOWEL_LENGTH_MALUS,
                params.RATING_LENGTH_SAME_PHONEM,
                params.RATING_LENGTH_SAME_WORD,
                params.SCORE_SAME_TRANSCRIPTION,
                params.SCORE_SILENCE_AMPLITUDE,
                params.SCORE_DURATION,
                params.RATING_SPECTRAL_SIMILARITY,
                params.RATING_AMPLITUDE_DIFFERENCE,
                params.SCORE_SAME_AUDIO_WORD,
                params.RANDOM_SPAN,
            ]
        )
        / FACTOR
    )


def change_parameters(x):
    x = x * FACTOR
    params.MAXIMAL_MINIMAL_PHONEM_LENGH_MALUS = x[0]
    params.MAXIMAL_MAXIMAL_CONSONANT_LENGTH_MALUS = x[1]
    params.MAXIMAL_MAXIMAL_VOWEL_LENGTH_MALUS = x[2]
    params.RATING_LENGTH_SAME_PHONEM = x[3]
    params.RATING_LENGTH_SAME_WORD = x[4]
    params.SCORE_SAME_TRANSCRIPTION = x[5]
    params.SCORE_SILENCE_AMPLITUDE = x[6]
    params.SCORE_DURATION = x[7]
    params.RATING_SPECTRAL_SIMILARITY = x[8]
    params.RATING_AMPLITUDE_DIFFERENCE = x[9]
    params.SCORE_SAME_AUDIO_WORD = x[10]
    params.RANDOM_SPAN = x[11]


def rate_sentence(videos, s):
    # TODO seed ?
    print(f'generating combos for "{s}"')
    combos = sm.process_sm(s, videos)
    print(f'rating "{s}"')
    rates = []

    for c in combos[:NB_COMBOS]:
        rate, wave = sentence_mixing.video_creator.audio.concat_segments(
            c.get_audio_phonems()
        )
        wave = (
            sentence_mixing.logic.audio_analysis.resample(
                np.sum(wave, axis=1).reshape((-1,)), rate, 16000
            )
            .astype("int16")
            .copy(order="C")
        )

        decoder = get_decoder()

        decoder.start_utt()  # begin utterance processing
        decoder.process_raw(
            wave, False, True
        )  # process audio data with recognition enabled (no_search = False), as a full utterance (full_utt = True)
        decoder.end_utt()  # stop utterance processing

        recognized_sentence = decoder.hyp().hypstr

        rates.append(sentence_distance(s, recognized_sentence))

    print(f"returning from {s}")
    return rates


def rate_parameters(x):
    videos = sm.get_videos(video_urls)
    change_parameters(x)

    rates = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures_rates = executor.map(
            rate_sentence, [videos] * NB_COMBOS, sentences
        )
        for r in futures_rates:
            rates.extend(r)
    res = sum(rates) / len(rates)
    print(x, res)
    return res


def translate(seq, dic):
    return "".join(dic[c] for c in seq)


def levenshtein_distance(a, b):
    phonems = set(a).union(b)
    dic = {p: chr(i) for i, p in enumerate(phonems)}

    return Levenshtein.distance(translate(a, dic), translate(b, dic))


def sentence_distance(original, recognized):
    # Asume that generate_compatible_dictionary() has been called
    original_phonems = stt_dict.sentence_to_phonem_list(original)
    recognized_phonems = stt_dict.sentence_to_phonem_list(recognized)

    d = levenshtein_distance(original_phonems, recognized_phonems)

    return d



import json

config_errmsg = {
    "align_exe": "the path to the Montreal align executable",
    "dict_path": "the path to the phonem dictionary",
    "dict_consonant_vowel_path": "the path to the consonant-voyel phonem classifier dict",
    "trained_model": "the path to the trained model",
    "lang": "the lang of the subtitles (used to convert numbers into plain text)",
    "folder": "the folder where the sub chunks will be stored",
    "stt_model_path": "path to speech to text model folder",
    "stt_full_dict_path": "path to speech to text original dictionary",
    "stt_tmp_dict_path": "path where the compatible dictionary should be generated",
}

config = None
config_path = None


def set_config_path(path):
    global config
    config_path = path

    try:
        with open(config_path, encoding="utf-8") as f:
            config = json.load(f)
    except EnvironmentError:
        raise FileNotFoundError(
            "No configuration file found. Please create and fill a 'config.json' file."
        )


def set_config_dict(config_dict):
    global config
    config = config_dict


def is_ready():
    return config is not None


def get_property(name):
    try:
        return config[name]
    except KeyError:
        raise KeyError(
            f"Please add the {name} property to your config.json containing {config_errmsg[name]}."
        )


def set_temp_property(name, value):
    global config
    if name in config:
        raise KeyError(f"Property {name} already exists")

    config[name] = value



import functools
import itertools

import sentence_mixing.logic.audio_analysis as audio_analysis
import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.text_parser as tp
import sentence_mixing.logic.utils as utils
from sentence_mixing.model.abstract import Sentence, Word


def get_last_phonem(audio):
    if isinstance(audio, Sentence):
        audio = audio.words[-1]
    if isinstance(audio, Word):
        audio = audio.phonems[-1]
    return audio


def get_last_vowel(associations):
    for a in associations:
        if a.audio_phonem.get_type() == "VOWEL":
            return a.audio_phonem
    return None


@functools.lru_cache(maxsize=None)
def step_3_audio_spectral_rating(last_vowel, audio_phonem):
    # TODO split by words
    score = 0
    if last_vowel is not None and audio_phonem.get_type() == "VOWEL":
        (rate, last_vowel_wave), (_, current_wave) = tuple(
            audio_analysis.resample_highest_rate(
                [last_vowel.get_wave(), audio_phonem.get_wave()]
            )
        )
        score += audio_analysis.cross_power_spectral_density_sum(
            current_wave, last_vowel_wave, rate
        )
    return score


def step_3_audio_amplitude_rating(previous_associations, current_association):
    previous = list(
        itertools.islice(
            (
                asso.audio_phonem
                for asso in previous_associations
                if asso.target_phonem.transcription != "sp"
            ),
            params.AMPLITUDE_STEPS_BACK,
        )
    )
    if len(previous) > 0:
        return audio_analysis.rate_amplitude_similarity(
            previous, current_association.audio_phonem
        )
    else:
        return 0


def step_3_n_following_previous_phonems(associations):
    i = 0
    while i < len(associations) - 1:
        a = associations[i]
        parent = associations[i + 1]

        if a.audio_phonem.previous_in_seq() == parent.audio_phonem:
            i += 1
        else:
            break
    return i



import functools
import itertools

import sentence_mixing.logic.parameters as parameters
import sentence_mixing.logic.text_parser as tp
import sentence_mixing.model as model
from sentence_mixing.model.abstract import Word


def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def count_finite_iterable(iterable):
    return sum(1 for _ in iterable)


def get_phonems(words_or_phonems):
    return list(
        itertools.chain(
            *map(
                lambda w: w.phonems if isinstance(w, Word) else [w],
                words_or_phonems,
            )
        )
    )


def sequence_association(start_association, reverse=False):
    return map(
        lambda phs: model.association.association_builder(*phs),
        zip(
            sequence_phonem(start_association.target_phonem, reverse=reverse),
            sequence_phonem(start_association.audio_phonem, reverse=reverse),
        ),
    )


def sequence_phonem(phonem, reverse=False):
    while phonem is not None:
        yield phonem

        if not reverse:
            phonem = phonem.next_in_seq()
        else:
            phonem = phonem.previous_in_seq()


def sequence_word(word, force_first_blank=False):
    """
    This is a generator used to retrieve each word in a sentence, from the word given in parameter
    until the end of the sentence

    This function ignores <BLANK> token, which are artificial words added between each original
    words of a target sentence.

    Parameters:
    word - Word object where the sequence starts
    force_first_blank - used to force the function to return the first word if it is <BLANK>.
                        This is necessary to prevent first phonem of a word to 'steal' the <BLANK>.
                        Example:
                            - TargetSentence: "la <BLANK> puissance"
                            - In the subtitles, we can find full word "puissance"

                            If this system is not used, the best combo will associate phonem 'p'
                            (which is the first phonem of "puissance"), to word <BLANK> when
                            checking to the longest homophone sequence.
    """

    if force_first_blank and word.token == "<BLANK>":
        yield word

    while word is not None:
        if word.token != "<BLANK>":
            yield word
        word = word.next_in_seq()


@functools.lru_cache(maxsize=None)
def get_sequence_aligner_homophones(target_phonem, audio_phonem):
    w0, w1 = target_phonem.word, audio_phonem.word
    return list(
        itertools.takewhile(
            lambda ws: ws[0].has_same_phonems_transcription(ws[1]),
            zip(sequence_word(w0), sequence_word(w1),),
        )
    )


@functools.lru_cache(maxsize=None)
def are_homophones(target_word, audio_word):
    try:
        return tp.are_token_homophones(target_word.token, audio_word.token)
    except model.exceptions.TokenAmbiguityError:
        return target_word.has_same_phonems_transcription(audio_word)


@functools.lru_cache(maxsize=None)
def get_sequence_dictionary_homophones(
    target_phonem, audio_phonem, skip_force_blank=False
):
    """Retrieves the longest homophones sequence"""

    w0, w1 = target_phonem.word, audio_phonem.word
    return itertools.takewhile(
        lambda ws: are_homophones(*ws),
        # For audio word, we force the first <BLANK> token skipping to false
        # <BLANK> token words are only contained in token word
        zip(sequence_word(w0, skip_force_blank), sequence_word(w1, False),),
    )


def has_at_least_one_node(base_nodes, total, rating, steps_left):
    return (base_nodes - 1) * rating >= total * steps_left


def compute_nodes_left(base_nodes, total, rating):
    return (base_nodes - 1) * rating / total


def association_sequence_from_words(words):
    return map(
        lambda x: model.association.association_builder(*x),
        itertools.chain(
            *map(lambda w: zip(w[0].phonems, w[1].phonems), words)
        ),
    )


def recompute_scores(asso_scores, nodes_left, steps_left):
    sor = sorted(
        [(t, t[0].get_total_score() + sum(t[1])) for t in asso_scores],
        key=lambda x: x[1],
        reverse=True,
    )

    chosen = []
    total = 0
    modif = parameters.START_MODIF
    for asso_score, computed_rate in sor:
        computed_rate *= modif

        # Not enough nodes left
        if not has_at_least_one_node(
            nodes_left, total + computed_rate, computed_rate, steps_left
        ):
            # not trivial
            break

        # After each association pick, highly decreasesa score modifier
        modif /= parameters.RATE_POWER
        total += computed_rate
        chosen.append((asso_score[0], asso_score[1], computed_rate))
    # print(nodes_left, chosen)
    return chosen, total



import sys

import numpy as np
import scipy.optimize

import sentence_mixing.parameter_tuning.parameter_tweaking
import sentence_mixing.parameter_tuning.speech_to_text_dict as stt_dict
import sentence_mixing.sentence_mixer as sm


def optimize(max_iter, config_path):
    sm.prepare_sm(config_path)
    stt_dict.generate_compatible_dictionary()

    initial = (
        sentence_mixing.parameter_tuning.parameter_tweaking.get_parameters()
    )
    scipy.optimize.minimize(
        sentence_mixing.parameter_tuning.parameter_tweaking.rate_parameters,
        initial,
        bounds=[(0, 1)] * 12,
        options={"maxiter": max_iter, "eps": 1e-3, "disp": True},
        callback=lambda x: print(x),
    )



import math

import sentence_mixing.logic.global_audio_data as global_audio_data
import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.utils as utils
from sentence_mixing.model.choice import Choice, Combo
from sentence_mixing.model.exceptions import Interruption, PhonemError


def compute_children(target_phonem, nodes_left, choice):
    """
    This function choses several audio_phonems for a given target_phonem.
    Sequentially choses an association, then decreasing a score modifier.
    Once the score of the next association is too low, stops the function and returns the children.

    Argument:
    target_phonem - target phonem for which we try to find a convenient audio_phonem
    nodes_left - number of remaining nodes
                 When this number of resources is too low, stop exploring the children
    choice - choice that called this function
             None when selecting the root choices

    Returns:
    a list of children Choice corresponding to the selected associations

    Raises:
    PhonemError - the best scored association has a score of 0. The phonem was not found in audio
    """

    steps_left = target_phonem.word.sentence.get_phonem_size() - target_phonem.word.sentence.get_phonem_index(
        target_phonem
    )
    total = 0
    chosen = []

    # Get all associations, sorted by decreasing scores
    association_candidates = global_audio_data.get_candidates(target_phonem)

    # This modif will be used to attenuate score
    modif = params.START_MODIF

    for group in utils.grouper(
        association_candidates, params.ANALYSE_CHUNK_SIZE
    ):
        candidates = None
        if choice is not None:
            candidates = sorted(
                [
                    (asso, choice.compute_child_step_3_score(asso))
                    for asso in group
                    if asso is not None
                ],
                key=lambda x: x[0].get_total_score() + sum(x[1]),
                reverse=True,
            )
        else:
            candidates = zip(
                filter(lambda x: x is not None, group),
                [[0, 0, 0]] * params.ANALYSE_CHUNK_SIZE,
            )

        # Rating
        for association, step_3_rate in candidates:
            computed_rate = association.get_total_score()

            computed_rate += sum(step_3_rate)
            computed_rate *= modif

            # Not enough nodes left
            if not utils.has_at_least_one_node(
                nodes_left, total + computed_rate, computed_rate, steps_left
            ):
                # not trivial
                break

            # After each association pick, highly decreasesa score modifier
            modif /= params.RATE_POWER
            total += computed_rate
            chosen.append((association, step_3_rate, computed_rate))
        else:
            # if didn't break, don't break
            continue
        break

    # Total score is null: we assume that the desired phonem was not found in the audio subtitles
    if total == 0:
        raise PhonemError(target_phonem)

    chosen, total = utils.recompute_scores(chosen, nodes_left, steps_left)

    return [
        Choice(
            choice,
            association,
            utils.compute_nodes_left(nodes_left, total, computed_rate,),
            *rate
        )
        for association, rate, computed_rate in chosen
    ]


def get_n_best_combos(sentence, videos, n=100, interrupt_callback=None):
    """
    Computes best n combos for a given sentence and set of videos

    All the algorithm bases on score values.

    Scores are assigned in three different steps:
    -Step 1: the audio phonems are scored individually
    -Step 2: the target phonem and audio phonem associations are score individually
    -Step 3: an association is scored comparatively to all previous chosen associations

    Steps 1 and 2 are computed exhaustively.
    However, step 3 cannot be computed exchaustively: the number of association combos is well
    to high.

    To counter this problem, we use a limited number of "nodes" Choices.
    """

    global_audio_data.set_videos(videos)

    target_phonems = utils.get_phonems(sentence.words)

    roots = compute_children(target_phonems[0], params.NODES, None)

    combos = []
    for combo in roots:
        if interrupt_callback is not None:
            if interrupt_callback():
                raise Interruption(interrupt_callback)

        combos.extend(combo.get_combos())

    combos.sort(key=lambda c: c.get_score(), reverse=True)

    return combos[:n]



import sentence_mixing.logic.analyze_step_1 as step_1
import sentence_mixing.logic.randomizer as rnd
from sentence_mixing.model.abstract import Phonem, Sentence, Word
from sentence_mixing.model.scorable import Scorable


class VideoSegment:
    """
    This class represents an abstract video segment
    All the following audio classes are inherited from this object, soidentifies precise spots of
    the video

    Internal attributes:
    - start: start timestamp of the segment in the audio file
    - end: end timestamp of the segment in the audio file
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def _get_original_wave(self):
        raise NotImplementedError()

    def get_wave(self):
        rate, data = self._get_original_wave()
        return rate, data[int(self.start * rate) : int(self.end * rate)]

    def _get_original_video(self):
        raise NotImplementedError()

    def get_video_clip(self):
        clip = self._get_original_video()
        return clip.subclip(self.start, self.end)

    def get_length(self):
        return self.end - self.start

    def __repr__(self):
        return f"VideoSegment({self.start}, {self.end}, {type(self)})"


class SubtitleLine(Sentence, VideoSegment):
    """Represents a line of subtitle in a video"""

    def __init__(self, original_text, start, end, video):
        Sentence.__init__(self, original_text)
        VideoSegment.__init__(self, start, end)
        self.video = video
        self.words = []

    def _get_original_wave(self):
        return self.video.get_audio_wave()

    def _get_original_video(self):
        return self.video.get_video()

    def get_index_in_video(self):
        return self.video.get_index_subtitle(self)

    def next_in_seq(self):
        if self == self.video.subtitles[-1]:
            return None
        return self.video.subtitles[self.get_index_in_video() + 1]

    def previous_in_seq(self):
        if self == self.video.subtitles[0]:
            return None
        return self.video.subtitles[self.get_index_in_video() - 1]


class AudioWord(Word, VideoSegment):
    """Represents a word spotted by Montreal aligner in a sentence"""

    def __init__(self, subtitle, token, original_word, start, end):
        Word.__init__(self, SubtitleLine, subtitle, token, original_word)
        VideoSegment.__init__(self, start, end)

    def _get_original_wave(self):
        return self.sentence._get_original_wave()

    def _get_original_video(self):
        return self.sentence._get_original_video()


class AudioPhonem(Phonem, VideoSegment, Scorable):
    """
    Represents a phonem spotted by Montreal aligner in a sentence
    An AudioPhonem has a step 1 phonem score.
    """

    def __init__(self, word, transcription, start, end):
        Phonem.__init__(self, AudioWord, word, transcription)
        VideoSegment.__init__(self, start, end)

        self.random_default_score = rnd.random_basic_score()
        self.length_score = step_1.score_length(self)

    def _get_original_wave(self):
        return self.word._get_original_wave()

    def _get_original_video(self):
        return self.word._get_original_video()

    def _get_split_score(self):
        assert self.random_default_score is not None
        assert self.length_score is not None

        return {
            "step_1_random": self.random_default_score,
            "step_1_duration": self.length_score,
        }



import os
import re

import youtube_dl


# workaround for youtube-dl issue #5710
class Logger(object):
    final_path = None

    def debug(self, msg):
        match = re.search(r"^\s*\[download\]\s(.*?)\shas.*$", msg)
        if match:
            self.final_path = match.group(1)

    def warning(self, msg):
        pass

    def error(self, msg):
        print(msg)


def dl_video(url):
    log = Logger()
    ydl_opts = {
        "outtmpl": ".downloads/%(id)s.%(title)s.%(ext)s",
        "logger": log,
    }
    with youtube_dl.YoutubeDL(ydl_opts) as ydl:
        filename = ydl.prepare_filename(ydl.extract_info(url))
        if not os.path.exists(filename):
            ydl.download([url])
        return log.final_path or filename
